// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef VKF_HYPOTHESIS_HPP
#define VKF_HYPOTHESIS_HPP

#include "MCState.hpp" // includes Fragment.hpp, Parents.hpp, Encoded.hpp, Sample.hpp, and Candidate.hpp too
#include "Obstacles.hpp" // includes Fragment.hpp, Parents.hpp, Encoded.hpp, and Sample.hpp too

namespace vkf {
class Hypothesis {
    protected:
    private:
        Candidate m_Candidate; //!< associated Candidate object
        bool is_Suspicious; //!< suspicious status of the hypothesis
    public:
        /** default constructor */
        Hypothesis() = delete;
        /** copy-constructor */
        Hypothesis(const Hypothesis&) = delete;
        /** move-constructor */
        Hypothesis(Hypothesis&&) = default;
        /** copy-operator= */
        Hypothesis& operator=(const Hypothesis& rhs) = delete;
        /** move-operator= */
        Hypothesis& operator=(Hypothesis&& rhs) = default;
        /** \brief Hypothesis constructor implementation
         * \param context determines an AOMatrix object to generate Hypothesis with respect to
         * \param obstacle lists counterexamples to block Hypothesis */
        Hypothesis(const AOMatrix& context, const Obstacles& obstacles)  : is_Suspicious(false) {
            bool hasObstacle = true;
            do {
                MCState&& candidates_pair(context);
                while(!candidates_pair.IsCoupled()) {
                    candidates_pair.MoveBothComponents(context);
                }
//                std::cout << candidates_pair.GetUpperCandidate().GetAttributes() << std::endl;
                candidates_pair.Complete(context);
                if (candidates_pair.GetUpperCandidate().GetObjects().count() < 2) {
                    hasObstacle = true;
                }
                else {
                    hasObstacle = false;
                    for (Obstacles::const_iterator cit = obstacles.cbegin(); cit != obstacles.cend(); ++cit) {
                        if (candidates_pair.GetUpperCandidate().GetAttributes().is_subset_of(**cit)) {
                            hasObstacle = true;
                            break;
                        }
                    }
                }
                if (!hasObstacle) {
                    m_Candidate = std::move(candidates_pair.GetUpperCandidate());
                }
            } while(hasObstacle);
	    };

        Hypothesis(bool suspicious, boost::dynamic_bitset<>&& encoded_bitset)  : m_Candidate(std::move(encoded_bitset)), is_Suspicious(suspicious) {}

        /**< default destructor */
        ~Hypothesis() = default;
        /** Access to the fragment of the candidate
         * \return m_PointerToCandidate->GetAttributes */
        Fragment& GetFragment() {return m_Candidate.GetAttributes();}
//        Fragment& GetFragment() const {return m_Candidate->GetAttributes();} // error
        /** Access to number of parents
         * \return size of the list of parents */
//        std::size_t& GetParentsNumber() {return m_Candidate.GetObjects().count();} // error
        std::size_t GetParentsNumber() {return m_Candidate.GetObjects().count();}
        /** Find position of the first parent
         * \return boost::dynamic_bitset<>::size_type or boost::dynamic_bitset<>::npos */
        boost::dynamic_bitset<>::size_type FindFirst() {return m_Candidate.GetObjects().find_first();};
        /** Find position of parent next to given position
         * \param pos determines a position to search next parent with respect to
         * \return boost::dynamic_bitset<>::size_type or boost::dynamic_bitset<>::npos */
        boost::dynamic_bitset<>::size_type FindNext(boost::dynamic_bitset<>::size_type pos) {return m_Candidate.GetObjects().find_next(pos);};
        /** Detect does an object at given position be a parent of the hypothesis
         * \param pos determines a position to check
         * \return true, if an object at given position is a parent of the hypothesis */
        bool IsParent(boost::dynamic_bitset<>::size_type pos) {return m_Candidate.GetObjects()[pos];};

        bool IsSuspicious() { return is_Suspicious; };
        void SetSuspiciousStatus(bool new_status) {is_Suspicious = new_status; return;};
        void SetParents(Parents&& parents) {m_Candidate.SetParents(std::move(parents));}
        void SetFragment(Fragment&& fragment) {m_Candidate.SetFragment(std::move(fragment));}
};

} // namespace vkf
#endif // VKF_HYPOTHESIS_HPP
