// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef VKF_INHABIT_HPP
#define VKF_INHABIT_HPP

#include <algorithm>
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <cassert>
#include <cmath>
#include <list>
#include <utility>

#include "Join.hpp"

namespace vkf {

bool pair_compare( std::pair<double, bool> const &pair1, std::pair<double, bool> const &pair2 );

double entropy(std::size_t pos, std::size_t neg);

double gain(std::size_t left_pos, std::size_t left_neg, std::size_t right_pos, std::size_t right_neg);

class Inhabit : public std::list<std::pair<double, bool>> {
    public:
        /** default destructor */
        ~Inhabit() = default;
        /** default constructor */
        Inhabit() = default;
        Inhabit(const vkf::Join &extended, std::uint32_t pos) {
            std::uint32_t sample_size = extended.size();
            for (std::uint32_t ndx = 0; ndx < sample_size; ndx++) {
                emplace_back(extended[ndx][pos],extended[ndx].IsPlus());
            }
            this->sort(vkf::pair_compare);
        };

        /** ComputeTheshold detects the best entropy verge for sorted values of continuous attribute
         * \param begin_it sets the start point to search the best verge
         * \param end_it determines a position to stop the search (an element on this position is ignored)
         * \return double value of the best theshold */
        double ComputeVerge(std::list<std::pair<double, bool>>::iterator begin_it,
                            std::list<std::pair<double, bool>>::iterator end_it
                            );
        /** ComputeEntropy computes the entropy of the values from begin_it to end_it
         * \param begin_it sets the start point
         * \param end_it determines a position to stop the search
         * \return double value of the entropy from begin_it to end_it */
        double ComputeEntropy(  std::list<std::pair<double, bool>>::iterator begin_it,
                                std::list<std::pair<double, bool>>::iterator end_it
                                );
        /** ComputeLeftEntropy computes the entropy of the values from begin_it to the verge
         * \param begin_it sets the start point to search the best verge
         * \param end_it determines a position to stop the search (an element on this position is ignored)
         * \return double value of the entropy from begin_it to ComputeVerge() */
        double ComputeLeftEntropy(  std::list<std::pair<double, bool>>::iterator begin_it,
                                    std::list<std::pair<double, bool>>::iterator end_it
                                    );
        /** ComputeRightEntropy computes the entropy of the values from the verge to the end_it
         * \param begin_it sets the start point to search the best verge
         * \param end_it determines a position to stop the search (an element on this position is ignored)
         * \return double value of the entropy from ComputeVerge() to the end_it */
        double ComputeRightEntropy( std::list<std::pair<double, bool>>::iterator begin_it,
                                    std::list<std::pair<double, bool>>::iterator end_it
                                    );
};

} // namespace vkf

#endif // VKF_INHABIT_HPP
